package no.nav.eia.dynaregler.helpers;

import java.util.ArrayList;
import java.util.Date;

import no.nav.eia.dynaregler.element.ParameterElement;
import no.nav.eia.dynaregler.element.RegelElement;
import no.nav.eia.dynaregler.element.SykmeldingParameterElement;

import org.apache.log4j.Logger;

/**
 * 
 * @author Florian van Eizenga (E140074)
 * Denne klassen er en sttteklasse kun for  behandle dynamiske regler.
 * 
 */

public abstract class RuleExecutionHelper {

	private static final Logger infolog = Logger.getLogger("infolog");
	private static final Logger errlog = Logger.getLogger("errorlog");
	
	private static void logInfo(String entry) {
		
		Date date = new Date();
		infolog.info("**********");
		infolog.info(date.toString());
		infolog.info(entry + "\n");
	}
	
	private static void logError(String entry, Exception e) {
		
		Date date = new Date();		
		errlog.error("**********");
		errlog.error(date.toString());
		errlog.error(e.getMessage());
		errlog.error(entry + "\n", e);
	}
	
	/**
	 * evaluerReglerAndGetResultatList
	 * 
	 * Denne metoden fyrer alle reglene ls p et SykmeldingParameter-objekt (nkkelverdier av sykmeldingen).
	 * I yeblikket (13-12-2013) bestr en regel av tre parametere: hoveddiagnose, tidsgruppe og kjnn.
	 * Nr en regel slr til, blir den lagres i regelResultatList.
	 * 
	 * @param regelList
	 * @param sykmeldingParam
	 * @return liste av regler som er relevante for denne sykmeldingen
	 */
	public static ArrayList<RegelElement> evaluerReglerAndGetResultatList(ArrayList<RegelElement> regelList, SykmeldingParameterElement sykmeldingParam) {
		
		ArrayList<RegelElement> regelResultatList = new ArrayList<RegelElement>();
				
		for (RegelElement regel : regelList) {
					
			if (		(evaluerHovedDiagnoseParameter(regel, sykmeldingParam))
					&& 	(evaluerTidsgruppeParameter(regel, sykmeldingParam.getSyketilfelleStartDato()
							, sykmeldingParam.getSykmeldingFOMDato(), sykmeldingParam.getSykmeldingTOMDato()))
					&& 	(evaluerKjonnParameter(regel, sykmeldingParam))) {
			
				regel.setRule(true);
				
			} else {
				
				regel.setRule(false);
			}
			
			regelResultatList.add(regel);
		}
		
		return regelResultatList;
	}
	
	/**
	 * evaluerHovedDiagnoseParameter
	 * 
	 * Denne metoden tar diagnosegruppeparameteren (ICPC-2 og ICD10) av regelen,
	 * og bruker det til  kontrollere et SykmeldingParameter-objekt.
	 * En diagnoseGruppe kan inneholde en kode 'A03' eller et regEx '[ABDFHKLNPRSTUWXYZ][0-2][1-9]'.
	 * 
	 * @param regel
	 * @param sykmeldingParameter
	 * @return true nr diagnosekode av denne sykmeldingen korresponderer med en diagnosekodeparameter (til en dynaregel), 
	 * 		   eller diagnosekodeparameter (til en dynaregel) ikke er oppgitt, ellers false
	 */
	private static boolean evaluerHovedDiagnoseParameter(RegelElement regel, SykmeldingParameterElement sykmeldingParameter) {
		
		ArrayList<ParameterElement> regelParameterList = regel.getParameterList();
		
		boolean regelResult = false;
		
			for (ParameterElement regelParameter : regelParameterList) {
				
				regelResult = false;
				
				String regelDiagnoseGruppeICPC2 = regelParameter.getDiagGruppeICPC2();
				boolean emptyRegelDiagnoseGruppeICPC2 = false;
				
				String regelDiagnoseGruppeICD10 = regelParameter.getDiagGruppeICD10();
				boolean emptyRegelDiagnoseGruppeICD10 = false;
				
				String sykmeldingParamDiagnoseGruppeICPC2 = regelParameter.getDiagGruppeICPC2();
				boolean emptySykmeldingParamDiagnoseGruppeICPC2 = false;
				
				String sykmeldingParamDiagnoseGruppeICD10 = regelParameter.getDiagGruppeICD10();
				boolean emptySykmeldingParamDiagnoseGruppeICD10 = false;
				
				if (regelDiagnoseGruppeICPC2 == null || (regelDiagnoseGruppeICPC2.isEmpty())) {
					
					emptyRegelDiagnoseGruppeICPC2 = true;
				}
								
				if (regelDiagnoseGruppeICD10 == null || (regelDiagnoseGruppeICD10.isEmpty())) {
					
					emptyRegelDiagnoseGruppeICD10 = true;
				}
				
				if ((emptyRegelDiagnoseGruppeICPC2) && (emptyRegelDiagnoseGruppeICD10)) {
					
					regelResult = true;
				}
				
				if (sykmeldingParamDiagnoseGruppeICPC2 == null || (sykmeldingParamDiagnoseGruppeICPC2.isEmpty())) {
					
					emptySykmeldingParamDiagnoseGruppeICPC2 = true;
				}
								
				if (sykmeldingParamDiagnoseGruppeICD10 == null || (sykmeldingParamDiagnoseGruppeICD10.isEmpty())) {
					
					emptySykmeldingParamDiagnoseGruppeICD10 = true;
				}
				
				if ((emptySykmeldingParamDiagnoseGruppeICPC2) && (emptySykmeldingParamDiagnoseGruppeICD10)) {
					
					regelResult = true;
				}
				
				if (	   (!emptyRegelDiagnoseGruppeICPC2) 
						&& (!emptySykmeldingParamDiagnoseGruppeICPC2)
						&& (isSykmeldingParameterICPC2(sykmeldingParameter))) {
					
					if (regelParameter.getDiagGruppeICPC2().contains(sykmeldingParameter.getHovedDiagnose().getKodeV())) {
						
						regelResult = true;
						
					} else if (runRegExpression(regelParameter.getDiagGruppeICPC2(), sykmeldingParameter.getHovedDiagnose().getKodeV())) {
						
						regelResult = true;
					}
				} else if ((!emptyRegelDiagnoseGruppeICD10) 
						&& (!emptySykmeldingParamDiagnoseGruppeICD10) 
						&& isSykmeldingParameterICD10(sykmeldingParameter)) {
	
					if(regelParameter.getDiagGruppeICD10().contains(sykmeldingParameter.getHovedDiagnose().getKodeV())) {
						
						regelResult = true;
						
					} else if (runRegExpression(regelParameter.getDiagGruppeICD10(), sykmeldingParameter.getHovedDiagnose().getKodeV())) {
						
						regelResult = true;
					}
				}
			}
			
		return regelResult;
	}

	/**
	 * evaluerTidsgruppeParameter
	 * 
	 * Denne metoden tar tidsgruppeparameteren av regelen og bruker det til  kontrollere et SykmeldingParameter-objekt.
	 * En tidsgruppe kan inneholde et tall '56' eller et utrykk 'lt17'.
	 * Nr en av tidsgruppe finns med et tall eller et uttrykk p det, 
	 * blir perioder beregnet: bde regelperioder og sykmeldingsperioder.
	 * 
	 * @param regel
	 * @param stSykmeldingParameterSyketilfelleStartDato
	 * @param stSykmeldingParameterFOMDato
	 * @param stSykmeldingParameterSykmeldingTOMDato
	 * @return true nr tidsgruppe av denne sykmeldingen korresponderer med en tidsgruppeparameter (til en dynaregel) 
	 * 		   eller tidsgruppeparameter (til en dynaregel) ikke er oppgitt, ellers false
	 */
	private static boolean evaluerTidsgruppeParameter(RegelElement regel, Date sykmeldingParameterSyketilfelleStartDato
			, Date sykmeldingParameterFOMDato, Date sykmeldingParameterTOMDato) {
		
		ArrayList<ParameterElement> regelParameterList = regel.getParameterList();
		
		boolean regelResult = false;
	
		for (ParameterElement regelParameter : regelParameterList) {
			
			regelResult = false;
					
				String regelParameterTidsGruppe = regelParameter.getTidsGruppe();
				boolean emptyTidsGruppe = regelParameterTidsGruppe.isEmpty();
				
				if ((regelParameterTidsGruppe == null) || (regelParameterTidsGruppe.isEmpty())) {
					
					emptyTidsGruppe = true;					
					regelResult = true;
				}
				
				//A) Nr det er oppgitt et tall
				if ((!emptyTidsGruppe) 
						&& (!regelParameterTidsGruppe.startsWith("lt") 
						&& (!regelParameterTidsGruppe.startsWith("gt")) 
						&& (!regelParameterTidsGruppe.startsWith("lf")) 
						&& (!regelParameterTidsGruppe.startsWith("gt")))) {
										
					Date resultFomMinStartTilfelle = new Date(sykmeldingParameterFOMDato.getTime() - sykmeldingParameterSyketilfelleStartDato.getTime());
					long antallDagerFomMinStartTilfelle = resultFomMinStartTilfelle.getTime() / (1000 * 60 * 60 * 24);
					
					Date resultTomMinStartTilfelle = new Date(sykmeldingParameterTOMDato.getTime() - sykmeldingParameterSyketilfelleStartDato.getTime());
					long antallDagerTomMinStartTilfelle = resultTomMinStartTilfelle.getTime() / (1000 * 60 * 60 * 24);
					
					String[] tidsGruppePeriodeArray = regelParameterTidsGruppe.split(",");
					long tidsGruppePeriode = 0;
					
					for (int i = 0; i < tidsGruppePeriodeArray.length; i++) {
						
						try {
							tidsGruppePeriode = Long.parseLong(tidsGruppePeriodeArray[i]);
						}
						catch (NumberFormatException nfe) {}
						
						if ((antallDagerFomMinStartTilfelle <= tidsGruppePeriode) & (antallDagerTomMinStartTilfelle >= tidsGruppePeriode)) {
							
							regelResult = true;
							break;
						}
					}										
				}
				
				//B) Nr det er angitt lt
				if (regelParameterTidsGruppe.startsWith("lt")) {
					
					String tall = regelParameterTidsGruppe.substring(2, regelParameterTidsGruppe.length());
					long tidsGruppePeriode = Long.parseLong(tall);
					
					Date result = new Date(sykmeldingParameterTOMDato.getTime() - sykmeldingParameterSyketilfelleStartDato.getTime());
					long antallDager = result.getTime() / (1000 * 60 * 60 * 24);
					
					if (antallDager < tidsGruppePeriode) {
						
						regelResult = true;
					}
				}
				
				//C) Nr det er angitt gt (strre enn tom dato) foran tallet
				if (regelParameterTidsGruppe.startsWith("gt")) {
					
					String tall = regelParameterTidsGruppe.substring(2, regelParameterTidsGruppe.length());
					long tidsGruppePeriode = Long.parseLong(tall);
					
					Date result = new Date(sykmeldingParameterTOMDato.getTime() - sykmeldingParameterSyketilfelleStartDato.getTime());
					long antallDager = result.getTime() / (1000 * 60 * 60 * 24);
					
					if (antallDager > tidsGruppePeriode) {
						
						regelResult = true;
					}
				}
				
				//D) Nr det er angitt lf (mindre enn fom dato) foran tallet
				//   (SykmeldingFOMDato - SyketilfelleStartDato) < tidspunkt
				if (regelParameterTidsGruppe.startsWith("lf")) {
					
					String tall = regelParameterTidsGruppe.substring(2, regelParameterTidsGruppe.length());
					long tidsGruppePeriode = Long.parseLong(tall);
					
					Date result = new Date(sykmeldingParameterFOMDato.getTime() - sykmeldingParameterSyketilfelleStartDato.getTime());
					long antallDager = result.getTime() / (1000 * 60 * 60 * 24);
					
					if (antallDager < tidsGruppePeriode) {
						
						regelResult = true;
					}
				}
				
				//E) Nr det er angitt gf (mindre enn fom dato) foran tallet
				//   (SykmeldingFOMDato - SyketilfelleStartDato) > tidspunkt
				if (regelParameterTidsGruppe.startsWith("gt")) {
					
					String tall = regelParameterTidsGruppe.substring(2, regelParameterTidsGruppe.length());
					long tidsGruppePeriode = Long.parseLong(tall);
					
					Date result = new Date(sykmeldingParameterFOMDato.getTime() - sykmeldingParameterSyketilfelleStartDato.getTime());
					long antallDager = result.getTime() / (1000 * 60 * 60 * 24);
					
					if (antallDager > tidsGruppePeriode) {
						
						regelResult = true;
					}
				}
		}
		
		return regelResult;
	}

	/**
	 * evaluerKjonnParameter
	 * 
	 * Denne metoden tar kjnnparameteren av regelen og bruker det til  kontrollere et SykmeldingParameter objekt.
	 * Et kjnn kan inneholde et stor bokstav 'K' eller 'M'.
	 * 
	 * @param regel
	 * @param sykmeldingParameter
	 * @return true nr kjnn av denne sykmeldingen korresponderer med en kjnnparameter (til en dynaregel) 
	 * 		   eller kjnnparameter (til en dynaregel) ikke er oppgitt, ellers false
	 */
	private static boolean evaluerKjonnParameter(RegelElement regel, SykmeldingParameterElement sykmeldingParameter) {
	
		ArrayList<ParameterElement> regelParameterList = regel.getParameterList();
		
		boolean regelResult = false;
		
		for (ParameterElement regelParameter : regelParameterList) {
		
			String kjonn = regelParameter.getKjonn();
			boolean emptyKjonn = kjonn.isEmpty();
			
			if (kjonn == null || kjonn.isEmpty()) {
				
				emptyKjonn = true;					
				regelResult = true;
			}
		
			if ((!emptyKjonn) && (regelParameter.getKjonn().equals(sykmeldingParameter.getKjonn()))) {
				
				regelResult = true;
			}
		}
		
		return regelResult;
	}

	/**
	 * isSykmeldingParameterICPC2
	 * 
	 * Denne metoden vurderer om ICPC-2 kodeverket finns p et SykmeldingParameter-objekt.
	 * ICPC-2 = "2.16.578.1.12.4.1.1.7170"
	 * 
	 * @param sykmeldingParameter
	 * @return true nr kodesystemS av sykmeldingen korresponderer med denne metodens kodesystemS's verdi, ellers false
	 */
	private static boolean isSykmeldingParameterICPC2(SykmeldingParameterElement sykmeldingParameter) {
		
		String sykmeldingParameterKodeSystemS = sykmeldingParameter.getHovedDiagnose().getKodeSystemS();
		boolean emptysykmeldingParameterKodeSystemS = false;
		
		if (sykmeldingParameterKodeSystemS == null || sykmeldingParameterKodeSystemS.isEmpty()) {
			
			emptysykmeldingParameterKodeSystemS = true;
		}
		
		if ((!emptysykmeldingParameterKodeSystemS) && (sykmeldingParameterKodeSystemS.equals("2.16.578.1.12.4.1.1.7170"))) {
			
			return true;
		}
		else {
			
			return false;
		}
	}

	/**
	 * isSykmeldingParameterICD10
	 * 
	 * Denne metoden vurderer om ICD-10 kodeverket finns p et SykmeldingParameter-objekt
	 * ICD-10 = "2.16.578.1.12.4.1.1.7110"
	 * 
	 * @param sykmeldingParameter
	 * @return true nr kodesystemS av sykmeldingen korresponderer med denne metodens kodesystemS's verdi, ellers false
	 */
	private static boolean isSykmeldingParameterICD10(SykmeldingParameterElement sykmeldingParameter) {
		
		String sykmeldingParameterKodeSystemS = sykmeldingParameter.getHovedDiagnose().getKodeSystemS();
		boolean emptysykmeldingParameterKodeSystemS = false;
		
		if (sykmeldingParameterKodeSystemS == null || sykmeldingParameterKodeSystemS.isEmpty()) {
			
			emptysykmeldingParameterKodeSystemS = true;
		}
		
		if ((!emptysykmeldingParameterKodeSystemS) && (sykmeldingParameterKodeSystemS.equals("2.16.578.1.12.4.1.1.7110"))) {
			
			return true;
		}
		else {
			
			return false;
		}
	}

	private static boolean runRegExpression(String expression, String target) {
		
		return target.matches(expression);
	}
}